@@ -10,12 +10,15 @@ from datetime import timezone
 from async_generator import aclosing
 from dateutil.parser import parse as parse_date
 from sqlalchemy import func
+from sqlalchemy import or_
 from tornado import web
 from tornado.iostream import StreamClosedError
 
 from .. import orm
+from .. import scopes
+from ..roles import assign_default_roles
+from ..scopes import needs_scope
 from ..user import User
-from ..utils import admin_only
 from ..utils import isoformat
 from ..utils import iterate_until
 from ..utils import maybe_future
@@ -31,15 +34,41 @@ class SelfAPIHandler(APIHandler):
 
     async def get(self):
         user = self.current_user
-        if user is None:
-            # whoami can be accessed via oauth token
-            user = self.get_current_user_oauth_token()
         if user is None:
             raise web.HTTPError(403)
+
+        _added_scopes = set()
         if isinstance(user, orm.Service):
-            model = self.service_model(user)
+            # ensure we have the minimal 'identify' scopes for the token owner
+            identify_scopes = scopes.identify_scopes(user)
+            get_model = self.service_model
         else:
-            model = self.user_model(user)
+            identify_scopes = scopes.identify_scopes(user.orm_user)
+            get_model = self.user_model
+
+        # ensure we have permission to identify ourselves
+        # all tokens can do this on this endpoint
+        for scope in identify_scopes:
+            if scope not in self.expanded_scopes:
+                _added_scopes.add(scope)
+                self.expanded_scopes.add(scope)
+        if _added_scopes:
+            # re-parse with new scopes
+            self.parsed_scopes = scopes.parse_scopes(self.expanded_scopes)
+
+        model = get_model(user)
+
+        # add session_id associated with token
+        # added in 2.0
+        token = self.get_token()
+        if token:
+            model["session_id"] = token.session_id
+        else:
+            model["session_id"] = None
+
+        # add scopes to identify model,
+        # but not the scopes we added to ensure we could read our own model
+        model["scopes"] = sorted(self.expanded_scopes.difference(_added_scopes))
         self.write(json.dumps(model))
 
 
@@ -52,9 +81,10 @@ class UserListAPIHandler(APIHandler):
         user = self.users[orm_user]
         return any(spawner.ready for spawner in user.spawners.values())
 
-    @admin_only
+    @needs_scope('list:users')
     def get(self):
         state_filter = self.get_argument("state", None)
+        offset, limit = self.get_api_pagination()
 
         # post_filter
         post_filter = None
@@ -95,14 +125,53 @@ class UserListAPIHandler(APIHandler):
             # no filter, return all users
             query = self.db.query(orm.User)
 
-        data = [
-            self.user_model(u, include_servers=True, include_state=True)
-            for u in query
-            if (post_filter is None or post_filter(u))
-        ]
+        sub_scope = self.parsed_scopes['list:users']
+        if sub_scope != scopes.Scope.ALL:
+            if not set(sub_scope).issubset({'group', 'user'}):
+                # don't expand invalid !server=x filter to all users!
+                self.log.warning(
+                    "Invalid filter on list:user for {self.current_user}: {sub_scope}"
+                )
+                raise web.HTTPError(403)
+            filters = []
+            if 'user' in sub_scope:
+                filters.append(orm.User.name.in_(sub_scope['user']))
+            if 'group' in sub_scope:
+                filters.append(
+                    orm.User.groups.any(
+                        orm.Group.name.in_(sub_scope['group']),
+                    )
+                )
+
+            if len(filters) == 1:
+                query = query.filter(filters[0])
+            else:
+                query = query.filter(or_(*filters))
+
+        full_query = query
+        query = query.order_by(orm.User.id.asc()).offset(offset).limit(limit)
+
+        user_list = []
+        for u in query:
+            if post_filter is None or post_filter(u):
+                user_model = self.user_model(u)
+                if user_model:
+                    user_list.append(user_model)
+
+        total_count = full_query.count()
+        if self.accepts_pagination:
+            data = self.paginated_model(user_list, offset, limit, total_count)
+        else:
+            query_count = query.count()
+            if offset == 0 and total_count > query_count:
+                self.log.warning(
+                    f"Truncated user list in request that does not expect pagination. Processing {query_count} of {total_count} total users."
+                )
+            data = user_list
+
         self.write(json.dumps(data))
 
-    @admin_only
+    @needs_scope('admin:users')
     async def post(self):
         data = self.get_json_body()
         if not data or not isinstance(data, dict) or not data.get('usernames'):
@@ -142,15 +211,14 @@ class UserListAPIHandler(APIHandler):
             user = self.user_from_username(name)
             if admin:
                 user.admin = True
-                self.db.commit()
+            assign_default_roles(self.db, entity=user)
+            self.db.commit()
             try:
                 await maybe_future(self.authenticator.add_user(user))
             except Exception as e:
                 self.log.error("Failed to create user: %s" % name, exc_info=True)
                 self.users.delete(user)
-                raise web.HTTPError(
-                    400, "Failed to create user %s: %s" % (name, str(e))
-                )
+                raise web.HTTPError(400, f"Failed to create user {name}: {e}")
             else:
                 created.append(user)
 
@@ -158,82 +226,73 @@ class UserListAPIHandler(APIHandler):
         self.set_status(201)
 
 
-def admin_or_self(method):
-    """Decorator for restricting access to either the target user or admin"""
-
-    def m(self, name, *args, **kwargs):
-        current = self.current_user
-        if current is None:
-            raise web.HTTPError(403)
-        if not (current.name == name or current.admin):
-            raise web.HTTPError(403)
-
-        # raise 404 if not found
-        if not self.find_user(name):
-            raise web.HTTPError(404)
-        return method(self, name, *args, **kwargs)
-
-    return m
-
-
 class UserAPIHandler(APIHandler):
-    @admin_or_self
-    async def get(self, name):
-        user = self.find_user(name)
-        model = self.user_model(
-            user, include_servers=True, include_state=self.current_user.admin
-        )
+    @needs_scope(
+        'read:users',
+        'read:users:name',
+        'read:servers',
+        'read:users:groups',
+        'read:users:activity',
+        'read:roles:users',
+    )
+    async def get(self, user_name):
+        user = self.find_user(user_name)
+        if user is None:
+            raise web.HTTPError(404)
+        model = self.user_model(user)
         # auth state will only be shown if the requester is an admin
         # this means users can't see their own auth state unless they
         # are admins, Hub admins often are also marked as admins so they
         # will see their auth state but normal users won't
-        requester = self.current_user
-        if requester.admin:
+        if 'auth_state' in model:
             model['auth_state'] = await user.get_auth_state()
         self.write(json.dumps(model))
 
-    @admin_only
-    async def post(self, name):
+    @needs_scope('admin:users')
+    async def post(self, user_name):
         data = self.get_json_body()
-        user = self.find_user(name)
+        user = self.find_user(user_name)
         if user is not None:
-            raise web.HTTPError(409, "User %s already exists" % name)
+            raise web.HTTPError(409, "User %s already exists" % user_name)
 
-        user = self.user_from_username(name)
+        user = self.user_from_username(user_name)
         if data:
             self._check_user_model(data)
             if 'admin' in data:
                 user.admin = data['admin']
-                self.db.commit()
+                assign_default_roles(self.db, entity=user)
+        self.db.commit()
 
         try:
             await maybe_future(self.authenticator.add_user(user))
         except Exception:
-            self.log.error("Failed to create user: %s" % name, exc_info=True)
+            self.log.error("Failed to create user: %s" % user_name, exc_info=True)
             # remove from registry
             self.users.delete(user)
-            raise web.HTTPError(400, "Failed to create user: %s" % name)
+            raise web.HTTPError(400, "Failed to create user: %s" % user_name)
 
         self.write(json.dumps(self.user_model(user)))
         self.set_status(201)
 
-    @admin_only
-    async def delete(self, name):
-        user = self.find_user(name)
+    @needs_scope('delete:users')
+    async def delete(self, user_name):
+        user = self.find_user(user_name)
         if user is None:
             raise web.HTTPError(404)
         if user.name == self.current_user.name:
             raise web.HTTPError(400, "Cannot delete yourself!")
         if user.spawner._stop_pending:
             raise web.HTTPError(
-                400, "%s's server is in the process of stopping, please wait." % name
+                400,
+                "%s's server is in the process of stopping, please wait." % user_name,
             )
         if user.running:
             await self.stop_single_user(user)
             if user.spawner._stop_pending:
                 raise web.HTTPError(
                     400,
-                    "%s's server is in the process of stopping, please wait." % name,
+                    "%s's server is in the process of stopping, please wait."
+                    % user_name,
                 )
 
         await maybe_future(self.authenticator.delete_user(user))
@@ -245,14 +304,14 @@ class UserAPIHandler(APIHandler):
 
         self.set_status(204)
 
-    @admin_only
-    async def patch(self, name):
-        user = self.find_user(name)
+    @needs_scope('admin:users')
+    async def patch(self, user_name):
+        user = self.find_user(user_name)
         if user is None:
             raise web.HTTPError(404)
         data = self.get_json_body()
         self._check_user_model(data)
-        if 'name' in data and data['name'] != name:
+        if 'name' in data and data['name'] != user_name:
             # check if the new name is already taken inside db
             if self.find_user(data['name']):
                 raise web.HTTPError(
@@ -264,6 +323,8 @@ class UserAPIHandler(APIHandler):
                 await user.save_auth_state(value)
             else:
                 setattr(user, key, value)
+                if key == 'admin':
+                    assign_default_roles(self.db, entity=user)
         self.db.commit()
         user_ = self.user_model(user)
         user_['auth_state'] = await user.get_auth_state()
@@ -273,15 +334,14 @@ class UserAPIHandler(APIHandler):
 class UserTokenListAPIHandler(APIHandler):
     """API endpoint for listing/creating tokens"""
 
-    @admin_or_self
-    def get(self, name):
+    @needs_scope('read:tokens')
+    def get(self, user_name):
         """Get tokens for a given user"""
-        user = self.find_user(name)
+        user = self.find_user(user_name)
         if not user:
-            raise web.HTTPError(404, "No such user: %s" % name)
+            raise web.HTTPError(404, "No such user: %s" % user_name)
 
         now = datetime.utcnow()
-
         api_tokens = []
 
         def sort_key(token):
@@ -295,19 +355,9 @@ class UserTokenListAPIHandler(APIHandler):
                 continue
             api_tokens.append(self.token_model(token))
 
-        oauth_tokens = []
-        # OAuth tokens use integer timestamps
-        now_timestamp = now.timestamp()
-        for token in sorted(user.oauth_tokens, key=sort_key):
-            if token.expires_at and token.expires_at < now_timestamp:
-                # exclude expired tokens
-                self.db.delete(token)
-                self.db.commit()
-                continue
-            oauth_tokens.append(self.token_model(token))
-        self.write(json.dumps({'api_tokens': api_tokens, 'oauth_tokens': oauth_tokens}))
+        self.write(json.dumps({'api_tokens': api_tokens}))
 
-    async def post(self, name):
+    async def post(self, user_name):
         body = self.get_json_body() or {}
         if not isinstance(body, dict):
             raise web.HTTPError(400, "Body must be a JSON dict or empty")
@@ -335,23 +385,38 @@ class UserTokenListAPIHandler(APIHandler):
         if requester is None:
             # couldn't identify requester
             raise web.HTTPError(403)
-        user = self.find_user(name)
-        if requester is not user and not requester.admin:
-            raise web.HTTPError(403, "Only admins can request tokens for other users")
-        if not user:
-            raise web.HTTPError(404, "No such user: %s" % name)
-        if requester is not user:
-            kind = 'user' if isinstance(requester, User) else 'service'
+        self._jupyterhub_user = requester
+        self._resolve_roles_and_scopes()
+        user = self.find_user(user_name)
+        kind = 'user' if isinstance(requester, User) else 'service'
+        scope_filter = self.get_scope_filter('tokens')
+        if user is None or not scope_filter(user, kind):
+            raise web.HTTPError(
+                403,
+                f"{kind.title()} {user_name} not found or no permissions to generate tokens",
+            )
 
         note = body.get('note')
         if not note:
             note = "Requested via api"
             if requester is not user:
-                note += " by %s %s" % (kind, requester.name)
+                note += f" by {kind} {requester.name}"
 
-        api_token = user.new_api_token(
-            note=note, expires_in=body.get('expires_in', None)
-        )
+        token_roles = body.get('roles')
+        try:
+            api_token = user.new_api_token(
+                note=note,
+                expires_in=body.get('expires_in', None),
+                roles=token_roles,
+            )
+        except KeyError:
+            raise web.HTTPError(404, "Requested roles %r not found" % token_roles)
+        except ValueError:
+            raise web.HTTPError(
+                403,
+                "Requested roles %r cannot have higher permissions than the token owner"
+                % token_roles,
+            )
         if requester is not user:
             self.log.info(
                 "%s %s requested API token for %s",
@@ -366,6 +431,7 @@ class UserTokenListAPIHandler(APIHandler):
         token_model = self.token_model(orm.APIToken.find(self.db, api_token))
         token_model['token'] = api_token
         self.write(json.dumps(token_model))
+        self.set_status(201)
 
 
 class UserTokenAPIHandler(APIHandler):
@@ -377,45 +443,41 @@ class UserTokenAPIHandler(APIHandler):
         Raises 404 if not found for any reason
         (e.g. wrong owner, invalid key format, etc.)
         """
-        not_found = "No such token %s for user %s" % (token_id, user.name)
-        prefix, id = token_id[0], token_id[1:]
-        if prefix == 'a':
-            Token = orm.APIToken
-        elif prefix == 'o':
-            Token = orm.OAuthAccessToken
-        else:
+        not_found = f"No such token {token_id} for user {user.name}"
+        prefix, id_ = token_id[:1], token_id[1:]
+        if prefix != 'a':
             raise web.HTTPError(404, not_found)
         try:
-            id = int(id)
+            id_ = int(id_)
         except ValueError:
             raise web.HTTPError(404, not_found)
 
-        orm_token = self.db.query(Token).filter(Token.id == id).first()
+        orm_token = self.db.query(orm.APIToken).filter_by(id=id_).first()
         if orm_token is None or orm_token.user is not user.orm_user:
             raise web.HTTPError(404, "Token not found %s", orm_token)
         return orm_token
 
-    @admin_or_self
-    def get(self, name, token_id):
+    @needs_scope('read:tokens')
+    def get(self, user_name, token_id):
         """"""
-        user = self.find_user(name)
+        user = self.find_user(user_name)
         if not user:
-            raise web.HTTPError(404, "No such user: %s" % name)
+            raise web.HTTPError(404, "No such user: %s" % user_name)
         token = self.find_token_by_id(user, token_id)
         self.write(json.dumps(self.token_model(token)))
 
-    @admin_or_self
-    def delete(self, name, token_id):
+    @needs_scope('tokens')
+    def delete(self, user_name, token_id):
         """Delete a token"""
-        user = self.find_user(name)
+        user = self.find_user(user_name)
         if not user:
-            raise web.HTTPError(404, "No such user: %s" % name)
+            raise web.HTTPError(404, "No such user: %s" % user_name)
         token = self.find_token_by_id(user, token_id)
         # deleting an oauth token deletes *all* oauth tokens for that client
-        if isinstance(token, orm.OAuthAccessToken):
-            client_id = token.client_id
+        client_id = token.client_id
+        if token.client_id != "jupyterhub":
             tokens = [
-                token for token in user.oauth_tokens if token.client_id == client_id
+                token for token in user.api_tokens if token.client_id == client_id
             ]
         else:
             tokens = [token]
@@ -429,9 +491,14 @@ class UserTokenAPIHandler(APIHandler):
 class UserServerAPIHandler(APIHandler):
     """Start and stop single-user servers"""
 
-    @admin_or_self
-    async def post(self, name, server_name=''):
-        user = self.find_user(name)
+    @needs_scope('servers')
+    async def post(self, user_name, server_name=''):
+        user = self.find_user(user_name)
+        if user is None:
+            # this can be reached if a token has `servers`
+            # permission on *all* users
+            raise web.HTTPError(404)
+
         if server_name:
             if not self.allow_named_servers:
                 raise web.HTTPError(400, "Named servers are not enabled.")
@@ -445,7 +512,7 @@ class UserServerAPIHandler(APIHandler):
                         400,
                         "User {} already has the maximum of {} named servers."
                         "  One must be deleted before a new server can be created".format(
-                            name, self.named_server_limit_per_user
+                            user_name, self.named_server_limit_per_user
                         ),
                     )
         spawner = user.spawners[server_name]
@@ -455,7 +522,7 @@ class UserServerAPIHandler(APIHandler):
             self.set_status(202)
             return
         elif pending:
-            raise web.HTTPError(400, "%s is pending %s" % (spawner._log_name, pending))
+            raise web.HTTPError(400, f"{spawner._log_name} is pending {pending}")
 
         if spawner.ready:
             # include notify, so that a server that died is noticed immediately
@@ -474,9 +541,9 @@ class UserServerAPIHandler(APIHandler):
         self.set_header('Content-Type', 'text/plain')
         self.set_status(status)
 
-    @admin_or_self
-    async def delete(self, name, server_name=''):
-        user = self.find_user(name)
+    @needs_scope('delete:servers')
+    async def delete(self, user_name, server_name=''):
+        user = self.find_user(user_name)
         options = self.get_json_body()
         remove = (options or {}).get('remove', False)
 
@@ -501,7 +568,7 @@ class UserServerAPIHandler(APIHandler):
                 raise web.HTTPError(400, "Named servers are not enabled.")
             if server_name not in user.orm_spawners:
                 raise web.HTTPError(
-                    404, "%s has no server named '%s'" % (name, server_name)
+                    404, f"{user_name} has no server named '{server_name}'"
                 )
         elif remove:
             raise web.HTTPError(400, "Cannot delete the default server")
@@ -519,7 +586,7 @@ class UserServerAPIHandler(APIHandler):
         if spawner.pending:
             raise web.HTTPError(
                 400,
-                "%s is pending %s, please wait" % (spawner._log_name, spawner.pending),
+                f"{spawner._log_name} is pending {spawner.pending}, please wait",
             )
 
         stop_future = None
@@ -547,19 +614,19 @@ class UserAdminAccessAPIHandler(APIHandler):
     This handler sets the necessary cookie for an admin to login to a single-user server.
     """
 
-    @admin_only
-    def post(self, name):
+    @needs_scope('servers')
+    def post(self, user_name):
         self.log.warning(
             "Deprecated in JupyterHub 0.8."
             " Admin access API is not needed now that we use OAuth."
         )
         current = self.current_user
         self.log.warning(
-            "Admin user %s has requested access to %s's server", current.name, name
+            "Admin user %s has requested access to %s's server", current.name, user_name
         )
         if not self.settings.get('admin_access', False):
             raise web.HTTPError(403, "admin access to user servers disabled")
-        user = self.find_user(name)
+        user = self.find_user(user_name)
         if user is None:
             raise web.HTTPError(404)
 
@@ -574,7 +641,7 @@ class SpawnProgressAPIHandler(APIHandler):
 
     async def send_event(self, event):
         try:
-            self.write('data: {}\n\n'.format(json.dumps(event)))
+            self.write(f'data: {json.dumps(event)}\n\n')
             await self.flush()
         except StreamClosedError:
             self.log.warning("Stream closed while handling %s", self.request.uri)
@@ -603,12 +670,12 @@ class SpawnProgressAPIHandler(APIHandler):
 
             await asyncio.wait([self._finish_future], timeout=self.keepalive_interval)
 
-    @admin_or_self
-    async def get(self, username, server_name=''):
+    @needs_scope('read:servers')
+    async def get(self, user_name, server_name=''):
         self.set_header('Cache-Control', 'no-cache')
         if server_name is None:
             server_name = ''
-        user = self.find_user(username)
+        user = self.find_user(user_name)
         if user is None:
             # no such user
             raise web.HTTPError(404)
@@ -628,7 +695,7 @@ class SpawnProgressAPIHandler(APIHandler):
         ready_event = {
             'progress': 100,
             'ready': True,
-            'message': "Server ready at {}".format(url),
+            'message': f"Server ready at {url}",
             'html_message': 'Server ready at <a href="{0}">{0}</a>'.format(url),
             'url': url,
         }
@@ -647,7 +714,12 @@ class SpawnProgressAPIHandler(APIHandler):
             # check if spawner has just failed
             f = spawn_future
             if f and f.done() and f.exception():
-                failed_event['message'] = "Spawn failed: %s" % f.exception()
+                exc = f.exception()
+                message = getattr(exc, "jupyterhub_message", str(exc))
+                failed_event['message'] = f"Spawn failed: {message}"
+                html_message = getattr(exc, "jupyterhub_html_message", "")
+                if html_message:
+                    failed_event['html_message'] = html_message
                 await self.send_event(failed_event)
                 return
             else:
@@ -680,7 +752,12 @@ class SpawnProgressAPIHandler(APIHandler):
             # what happened? Maybe spawn failed?
             f = spawn_future
             if f and f.done() and f.exception():
-                failed_event['message'] = "Spawn failed: %s" % f.exception()
+                exc = f.exception()
+                message = getattr(exc, "jupyterhub_message", str(exc))
+                failed_event['message'] = f"Spawn failed: {message}"
+                html_message = getattr(exc, "jupyterhub_html_message", "")
+                if html_message:
+                    failed_event['html_message'] = html_message
             else:
                 self.log.warning(
                     "Server %s didn't start for unknown reason", spawner._log_name
@@ -731,7 +808,7 @@ class ActivityAPIHandler(APIHandler):
             if server_name not in spawners:
                 raise web.HTTPError(
                     400,
-                    "No such server '{}' for user {}".format(server_name, user.name),
+                    f"No such server '{server_name}' for user {user.name}",
                 )
             # check that each per-server field is a dict
             if not isinstance(server_info, dict):
@@ -746,12 +823,12 @@ class ActivityAPIHandler(APIHandler):
             )
         return servers
 
-    @admin_or_self
-    def post(self, username):
-        user = self.find_user(username)
+    @needs_scope('users:activity')
+    def post(self, user_name):
+        user = self.find_user(user_name)
         if user is None:
             # no such user
-            raise web.HTTPError(404, "No such user: %r", username)
+            raise web.HTTPError(404, "No such user: %r", user_name)
 
         body = self.get_json_body()
         if not isinstance(body, dict):
